#include "SystemSolverWHCAStar.h"

#include "SystemSolverCAStar.h"
#include "SystemSolver.h"
#include "ISystemSolver.h"
#include "ElectricCircuit.h"
#include "ScenarioData.h"
#include "SystemSolution.h"
#include "IVehicleSolver.h"
#include "PerformanceAnalysis.h"
#include "IArrangement.h"
#include "Vehicle.h"
#include "VehicleSolution.h"
#include <assert.h>
#include <iostream>
#include <algorithm>

SystemSolverWHCAStar::SystemSolverWHCAStar(IObjectiveFunction* inObjectiveFunction, IVehicleSolver* inVehicleSolver, IArrangement* inArrangment, ElectricCircuit* inElectricCircuit, float infWindowWidth, float infResolveStep)
	:SystemSolverArranged(inObjectiveFunction, inVehicleSolver, inArrangment, inElectricCircuit)
	, mfWindowWidth(infWindowWidth)
	, mfResolveStep(infResolveStep)
{
	
}


SystemSolverWHCAStar::~SystemSolverWHCAStar()
{
}

void SystemSolverWHCAStar::SolveSystem()
{
	if (!mIsActivated)
	{
		return;
	}

	//Init
	mPerformanceAnalysis->StartExperiment();
	ResetSolver();
	float fCurrentWindowStart = mfMinimumDepartureTime;
	float fCurrentWindowEnd = fCurrentWindowStart + mfWindowWidth;
	bool needToSolve = true;
	//Main loop. We iterate over the time windows
	while (needToSolve)
	{
		//Will yield true for first iteration
		needToSolve = mSystemSolution->mMapVehicleSolution.size() == 0;

		//Strip the current solution from actions that occur the start
		//Determine if we need to solve for next iteration
		//Will not occur for first iteration
		for each (auto var in mSystemSolution->mMapVehicleSolution)
		{
			VehicleSolution* newSolution = var.second;
			//Strip the solution from actions that started after time horizon
			bool wasChanged = newSolution->keepPartialSolution(fCurrentWindowStart);
			if (wasChanged || needToSolve || newSolution->actionCount() == 0)
			{
				needToSolve = true;
			}
		}

		//Cancel old solution validity
		GetCurrentSystemSolution()->mSystemObjectiveFunctionValue = INFINITY;
		//Create a new electric circuit from current state
		ElectricCircuit* simulatedCircuit = mSolverData->mElectricCircuit->clone();
		mVehicleSolver->ReplaceRefElectricCircuit(simulatedCircuit);
		SystemSolverCAStar* solverCAStar = new SystemSolverCAStar(mObjectiveFunction, mVehicleSolver, mArrangement, simulatedCircuit);
		ScenarioData* dataCAStar = solverCAStar->ModifyOptionnalScenarioData();
		dataCAStar->mfOptTimeHorizon = fCurrentWindowEnd;
		dataCAStar->mObjOptionnalPartialSolution = mSystemSolution->clone();
		//Add all vehicles, solver CA will take car of ordering
		for (unsigned int j = 0; j < mSolverData->mVecVehicle.size(); ++j)
		{
			solverCAStar->AddVehicle(mSolverData->mVecVehicle[j]);
		}
		solverCAStar->SolveSystem();

		//If we have a better solution keep the solver and order!
		if (solverCAStar->GetCurrentSystemSolution()->mSystemObjectiveFunctionValue < GetCurrentSystemSolution()->mSystemObjectiveFunctionValue)
		{
			ResetSolver();
			//mSystemSolution = solverImmediate->GetCurrentSystemSolution();
			mBestSolver = solverCAStar;
			mBestCircuit = simulatedCircuit;
		}
		else
		{
			//Otherwise, delete them
			delete solverCAStar;
			delete simulatedCircuit;
		}
		//Prepare next time
		fCurrentWindowStart += mfResolveStep;
		fCurrentWindowEnd    = fCurrentWindowStart + mfWindowWidth;
		mSystemSolution = GetCurrentSystemSolution();

		//Stop if you hit a wall
		needToSolve &= !mSystemSolution->IsImpossible();
	};

	//Add a desription when we have a solution
	if (mBestSolver != nullptr)
	{
		mBestSolver->GetCurrentSystemSolution()->mDescription = mSystemSolution->mDescription;
	}
	mPerformanceAnalysis->PauseExperiment(GetCurrentSystemSolution()->mSystemObjectiveFunctionValue);
}

void SystemSolverWHCAStar::AddVehicleInternal(Vehicle * newVehicle)
{
	mfMinimumDepartureTime = min(mfMinimumDepartureTime, newVehicle->GetDepartureTime());
	SystemSolverArranged::AddVehicleInternal(newVehicle);
}
